# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for the "main" driver of the special remotes"""

import logging
import os
import pytest

from ..main import main

from ..base import AnnexCustomRemote

from datalad.tests.utils_pytest import (
    patch,
    swallow_logs,
    swallow_outputs,
)
from datalad.ui import ui


def test_erroring_out():
    class TooAbstract(AnnexCustomRemote):
        pass

    # patch to not let `main` change the ui.backend
    # of the test process (instead of a special remote process it is actually
    # targeting)
    with swallow_logs(new_level=logging.DEBUG) as cml, \
        swallow_outputs() as cmo:
        with pytest.raises(SystemExit) as cme,\
                patch.object(ui, "set_backend", autospec=True):
            main(args=[], cls=TooAbstract)
        assert cme.value.code == 1
        assert 'passing ERROR to git-annex' in cml.out
        # verify basic correct formatting of string to git-annex
        assert cmo.out.startswith('ERROR ')
        assert os.linesep not in cmo.out.rstrip()
        assert cmo.out.endswith('\n')  # This is the case even on Windows.



from datalad.tests.utils_pytest import eq_

from ..renderer import (
    NA_STRING,
    nadict,
    nagen,
)


def test_nagen():
    na = nagen()
    eq_(str(na), NA_STRING)
    eq_(repr(na), 'nagen()')
    assert na.unknown is na
    assert na['unknown'] is na

    eq_(str(nagen('-')), '-')


def test_nadict():
    d = nadict({1: 2})
    eq_(d[1], 2)
    eq_(str(d[2]), NA_STRING)






# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""DataLad aims to expose (scientific) data available online as a unified data
distribution with the convenience of git-annex repositories as a backend.

Commands are exposed through both a command-line interface and a Python API. On
the command line, run 'datalad --help' for a summary of the available commands.
From an interactive Python session, import `datalad.api` and inspect its
documentation with `help`.
"""

if not __debug__:
    raise RuntimeError(
        'DataLad cannot run in "optimized" mode, i.e. python -O')

import atexit
import os

# this is not to be modified. for querying use get_apimode()
__api = 'python'


def get_apimode():
    """Returns the API mode label for the current session.

    The API mode label indicates whether DataLad is running in "normal"
    mode in a Python session, or whether it is used via the command line
    interface.

    This function is a utility for optimizing behavior and messaging to the
    particular API (Python vs command line) in use in a given process.

    Returns
    {'python', 'cmdline'}
      The API mode is 'python' by default, unless the main command line
      entrypoint set it to 'cmdline'.
    """
    return __api


def in_librarymode():
    """Returns whether DataLad is requested to run in "library mode"

    In this mode DataLad aims to behave without the assumption that it is
    itself the front-end of a process and in full control over messaging
    and parameters.

    Returns
    -------
    bool
    """
    return __runtime_mode == 'library'


def enable_librarymode():
    """Request DataLad to operate in library mode.

    This function should be executed immediately after importing the `datalad`
    package, when DataLad is not used as an application, or in interactive
    scenarios, but as a utility library inside other applications. Enabling
    this mode will turn off some convenience feature that are irrelevant in
    such use cases (with performance benefits), and alters it messaging
    behavior to better interoperate with 3rd-party front-ends.

    Library mode can only be enabled once. Switching it on and off within
    the runtime of a process is not supported.

    Example::

        >>> import datalad
        >>> datalad.enable_librarymode()
    """
    global __runtime_mode
    __runtime_mode = 'library'
    # export into the environment for child processes to inherit
    os.environ['DATALAD_RUNTIME_LIBRARYMODE'] = '1'


# For reproducible demos/tests
_seed = os.environ.get('DATALAD_SEED', None)
if _seed is not None:
    import random
    _seed = int(_seed)
    random.seed(_seed)

# Colorama (for Windows terminal colors) must be imported before we use/bind
# any sys.stdout
try:
    # this will fix the rendering of ANSI escape sequences
    # for colored terminal output on windows
    # it will do nothing on any other platform, hence it
    # is safe to call unconditionally
    import colorama
    colorama.init()
    atexit.register(colorama.deinit)
except ImportError as e:
    pass

# Other imports are interspersed with lgr.debug to ease troubleshooting startup
# delays etc.

from .config import ConfigManager

cfg = ConfigManager()


# must come after config manager
# this is not to be modified. see enable/in_librarymode()
# by default, we are in application-mode, simply because most of
# datalad was originally implemented with this scenario assumption
__runtime_mode = 'library' \
    if cfg.getbool('datalad.runtime', 'librarymode', False) \
    else 'application'

from datalad.utils import (
    get_encoding_info,
    get_envvars_info,
    getpwd,
)

from .log import lgr


def setup_package():
    import warnings

    warnings.warn(
        "setup_package() and testing with nose are deprecated."
        "  Switch to using pytest instead.",
        DeprecationWarning,
    )
    from datalad.tests.utils import setup_package
    return setup_package()


def teardown_package():
    import warnings
    warnings.warn(
        "teardown_package() and testing with nose are deprecated."
        "  Switch to using pytest instead.",
        DeprecationWarning,
    )
    from datalad.tests.utils import teardown_package
    return teardown_package()

# To analyze/initiate our decision making on what current directory to return
getpwd()

lgr.log(5, "Instantiating ssh manager")
from .support.sshconnector import SSHManager

ssh_manager = SSHManager()
atexit.register(ssh_manager.close, allow_fail=False)
atexit.register(lgr.log, 5, "Exiting")

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

if str(__version__) == '0' or __version__.startswith('0+'):
    lgr.warning(
        "DataLad was not installed 'properly' so its version is an uninformative %r.\n"
        "It can happen e.g. if datalad was installed via\n"
        "  pip install https://github.com/.../archive/{commitish}.zip\n"
        "instead of\n"
        "  pip install git+https://github.com/...@{commitish} .\n"
        "We advise to re-install datalad or downstream projects might not operate correctly.",
        __version__
    )

lgr.log(5, "Done importing main __init__")



"""Utilities and definitions for DataLad command interfaces"""

# TODO this should be a dochelper
from datalad.interface.base import dedent_docstring

# Some known extensions and their commands to suggest whenever lookup fails
_known_extension_commands = {
    'datalad-container': (
        'containers-list', 'containers-remove', 'containers-add',
        'containers-run'),
    'datalad-crawler': ('crawl', 'crawl-init'),
    'datalad-deprecated': (
        'ls',
        'metadata',
        'search',
        'aggregate-metadata',
        'extract-metadata',
    ),
    'datalad-neuroimaging': ('bids2scidata',)
}

_deprecated_commands = {
    'add': "save",
    'uninstall': 'drop',
}


def get_cmd_ex(interface):
    """Return the examples for the command defined by 'interface'.

    Parameters
    ----------
    interface : subclass of Interface
    """
    from datalad.interface.base import build_example
    intf_ex = "\n\n*Examples*\n\n"
    for example in interface._examples_:
        intf_ex += build_example(example, api='cmdline')
    return intf_ex


def get_cmdline_command_name(intfspec):
    """Given an interface specification return a cmdline command name"""
    if len(intfspec) > 2:
        name = intfspec[2]
    else:
        name = intfspec[0].split('.')[-1].replace('_', '-')
    return name


def alter_interface_docs_for_cmdline(docs):
    """Apply modifications to interface docstrings for cmdline doc use."""
    # central place to alter the impression of docstrings,
    # like removing Python API specific sections, and argument markup
    if not docs:
        return docs
    import re
    import textwrap

    docs = dedent_docstring(docs)
    # clean cmdline sections
    docs = re.sub(
        r'\|\| PYTHON \>\>.*?\<\< PYTHON \|\|',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # clean cmdline in-line bits
    docs = re.sub(
        r'\[PY:\s.*?\sPY\]',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    docs = re.sub(
        r'\[CMD:\s(.*?)\sCMD\]',
        lambda match: match.group(1),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    docs = re.sub(
        r'\|\| CMDLINE \>\>(.*?)\<\< CMDLINE \|\|',
        lambda match: match.group(1),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # remove :role:`...` RST markup for cmdline docs
    docs = re.sub(
        r':\S+:`[^`]*`[\\]*',
        lambda match: ':'.join(match.group(0).split(':')[2:]).strip('`\\'),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # make the handbook doc references more accessible
    # the URL is a redirect configured at readthedocs
    docs = re.sub(
        r'(handbook:[0-9]-[0-9]*)',
        '\\1 (http://handbook.datalad.org/symbols)',
        docs)
    # remove None constraint. In general, `None` on the cmdline means don't
    # give option at all, but specifying `None` explicitly is practically
    # impossible
    docs = re.sub(
        r',\sor\svalue\smust\sbe\s`None`',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # capitalize variables and remove backticks to uniformize with
    # argparse output
    docs = re.sub(
        r'([^`]+)`([a-zA-Z0-9_]+)`([^`]+)',
        lambda match: f'{match.group(1)}{match.group(2).upper()}{match.group(3)}',
        docs)
    # select only the cmdline alternative from argument specifications
    docs = re.sub(
        r'``([a-zA-Z0-9_,.]+)\|\|([a-zA-Z0-9-,.]+)``',
        lambda match: f'``{match.group(2)}``',
        docs)
    # clean up sphinx API refs
    docs = re.sub(
        r'\~datalad\.api\.\S*',
        lambda match: "`{0}`".format(match.group(0)[13:]),
        docs)
    # dedicated support for version markup
    docs = docs.replace('.. versionadded::', 'New in version')
    docs = docs.replace('.. versionchanged::', 'Changed in version')
    docs = docs.replace('.. deprecated::', 'Deprecated in version')
    # Remove RST paragraph markup
    docs = re.sub(
        r'^.. \S+::',
        lambda match: match.group(0)[3:-2].upper(),
        docs,
        flags=re.MULTILINE)
    docs = re.sub(
        r'^([ ]*)\|\| REFLOW \>\>\n(.*?)\<\< REFLOW \|\|',
        lambda match: textwrap.fill(match.group(2), subsequent_indent=match.group(1)),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    return docs



from __future__ import annotations

from typing import Optional
from unittest.mock import (
    call,
    patch,
)

from datalad.tests.utils_pytest import (
    assert_equal,
    assert_in,
    assert_is_none,
)

from ..utils import (
    AssemblingDecoderMixIn,
    LineSplitter,
)


test_lines = [
    "first line",
    "second line",
    "third line",
    ""
]


def _check_splitting_endings_separator(endings: list[str],
                                       separator: Optional[str] = None,
                                       keep_ends: bool = False,
                                       check_continuation: bool = False
                                       ):
    for line_ending in endings:
        line_splitter = LineSplitter(separator=separator, keep_ends=keep_ends)
        full_end = line_ending + separator if separator else line_ending
        if separator:
            expected_end = full_end if keep_ends else line_ending
        else:
            expected_end = line_ending if keep_ends else ""

        lines = line_splitter.process(
            full_end.join(test_lines)
            + full_end
            + ("fourth " if check_continuation else "")
        )
        assert_equal(
            lines,
            [line + expected_end for line in test_lines]
        )

        if check_continuation:
            assert_equal(line_splitter.remaining_data, "fourth ")
            lines = line_splitter.process("line" + full_end)
            assert_equal(
                lines,
                ["fourth line" + expected_end])
            assert_is_none(line_splitter.finish_processing())
        else:
            assert_is_none(line_splitter.finish_processing())


def test_line_splitter_basic():
    # expect lines without endings, split at standard line-endings
    _check_splitting_endings_separator(["\n", "\r\n"])
    _check_splitting_endings_separator(["\n", "\r\n"], check_continuation=True)


def test_line_splitter_basic_keep():
    # expect lines without endings, split at standard line-endings
    _check_splitting_endings_separator(["\n", "\r\n"], keep_ends=True)
    _check_splitting_endings_separator(
        ["\n", "\r\n"],
        keep_ends=True,
        check_continuation=True)


def test_line_splitter_zero():
    # expect lines without endings, split at standard line-endings
    _check_splitting_endings_separator(["\n", "\r\n"], separator="\x00")
    _check_splitting_endings_separator(
        ["\n", "\r\n"],
        separator="\x00",
        check_continuation=True)


def test_line_splitter_zero_keep():
    # expect lines without endings, split at standard line-endings
    _check_splitting_endings_separator(
        ["\n", "\r\n"],
        separator="\x00",
        keep_ends=True)
    _check_splitting_endings_separator(
        ["\n", "\r\n"],
        separator="\x00",
        keep_ends=True,
        check_continuation=True)


def test_line_splitter_corner_cases():
    line_splitter = LineSplitter()
    lines = line_splitter.process("")
    assert_equal(lines, [])
    assert_equal(line_splitter.remaining_data, None)

    line_splitter = LineSplitter()
    lines = line_splitter.process("")
    assert_equal(lines, [])
    lines = line_splitter.process("\n")
    assert_equal(lines, [""])
    assert_equal(line_splitter.remaining_data, None)

    line_splitter = LineSplitter()
    lines = line_splitter.process("  a   \f \r\n")
    assert_equal(lines, ["  a   ", " "])


def test_assembling_decoder_mix_in_basic():

    encoding = "utf-8"
    unicode_str = "These are not ASCII: ä, ö, ü. These can be ASCII: a, o, u."
    data_bytes = unicode_str.encode(encoding)

    adm = AssemblingDecoderMixIn()

    single_result = "".join([
        adm.decode(1, bytes([data_byte]), encoding)
        for data_byte in data_bytes
    ])
    assert_equal(single_result, unicode_str)


def _decode_multiple(adm: AssemblingDecoderMixIn,
                     encoded_strings: list[bytes],
                     encoding: str,
                     fixed_index: Optional[int] = None) -> list[str]:

    # Interleave decoding of multiple strings
    decoded_chars: list[list] = [list() for _ in range(len(encoded_strings))]
    for data_index in range(max([len(es) for es in encoded_strings])):
        for string_index in range(len(encoded_strings)):
            if data_index < len(encoded_strings[string_index]):
                decoded_char = adm.decode(
                    string_index if fixed_index is None else fixed_index,
                    bytes([encoded_strings[string_index][data_index]]),
                    encoding)
                decoded_chars[string_index].append(decoded_char)
    return ["".join(decoded_list) for decoded_list in decoded_chars]


def test_assembling_decoder_mix_in_multiple():
    encoding = "utf-8"
    unicode_strings = [
        "These are not ASCII: ä, ö, ü. These can be ASCII: a, o, u.",
        "Some other weird stuff: öäöß.",
        "Even weirder: 🐷🐶.",
    ]
    encoded_strings = [
        unicode_string.encode(encoding)
        for unicode_string in unicode_strings
    ]

    adm = AssemblingDecoderMixIn()
    decoded_strings = _decode_multiple(adm, encoded_strings, encoding)
    assert_equal(unicode_strings, decoded_strings)


def test_assembling_decoder_mix_in_multiple_fail():
    encoding = "utf-8"
    unicode_strings = [
        "A: ä, ö, ü.",
        "B: öäöß.",
        "C: 🐷🐶.",
    ]
    encoded_strings = [
        unicode_string.encode(encoding)
        for unicode_string in unicode_strings
    ]

    adm = AssemblingDecoderMixIn()
    decoded_strings = _decode_multiple(adm, encoded_strings, encoding, 0)
    # Because the strings are not separated, we do not expect any proper
    # output after single-byte encoded chars.
    assert_equal(decoded_strings, ["A: ", "B: ", "C: "])


def test_assembling_decoder_mix_in_warning():
    encoding = "utf-8"
    data_bytes = "🐷🐶.".encode(encoding)

    adm = AssemblingDecoderMixIn()

    with patch("datalad.runner.utils.logger") as logger_mock:
        result = adm.decode(1, data_bytes[0:1], encoding)
        assert_equal(result, '')
        del adm
        assert_in(
            call.warning("unprocessed data in AssemblingDecoderMixIn"),
            logger_mock.mock_calls)
        assert_in(
            call.debug(
                "unprocessed data in AssemblingDecoderMixIn:\n"
                "fd: 1, data: b'\\xf0'\n"),
            logger_mock.mock_calls)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""High-level interface for creating a publication target on a GOGS instance
"""

import logging
from urllib.parse import urlparse

from datalad.distributed.create_sibling_ghlike import (
    _create_sibling,
    _GitHubLike,
)
from datalad.distribution.dataset import datasetmethod
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)

lgr = logging.getLogger('datalad.distributed.create_sibling_gogs')


class _GOGS(_GitHubLike):
    """Customizations for the GOGS platform"""
    name = 'gogs'
    fullname = 'GOGS'
    create_org_repo_endpoint = 'api/v1/org/{organization}/repos'
    create_user_repo_endpoint = 'api/v1/user/repos'
    get_authenticated_user_endpoint = 'api/v1/user'
    get_repo_info_endpoint = 'api/v1/repos/{user}/{repo}'
    extra_remote_settings = {
        # first make sure that annex doesn't touch this one
        # but respect any existing config
        'annex-ignore': 'true',
    }

    def __init__(self, url, credential, require_token=True, token_info=None):
        if not url:
            raise ValueError(f'API URL required for {self.fullname}')
        return super().__init__(
            url,
            credential,
            require_token=require_token,
            token_info=f'Visit {url}/user/settings/applications '
                       'to create a token')


@build_doc
class CreateSiblingGogs(Interface):
    """Create a dataset sibling on a GOGS site

    GOGS is a self-hosted, free and open source code hosting solution with
    low resource demands that enable running it on inexpensive devices like
    a Raspberry Pi, or even directly on a NAS device.

    In order to be able to use this command, a personal access token has to be
    generated on the platform
    (Account->Your Settings->Applications->Generate New Token).

    This command can be configured with
    "datalad.create-sibling-ghlike.extra-remote-settings.NETLOC.KEY=VALUE" in
    order to add any local KEY = VALUE configuration to the created sibling in
    the local `.git/config` file. NETLOC is the domain of the Gogs instance to
    apply the configuration for.
    This leads to a behavior that is equivalent to calling datalad's
    ``siblings('configure', ...)``||``siblings configure`` command with the
    respective KEY-VALUE pair after creating the sibling.
    The configuration, like any other, could be set at user- or system level, so
    users do not need to add this configuration to every sibling created with
    the service at NETLOC themselves.

    .. versionadded:: 0.16
    """

    _params_ = _GOGS.create_sibling_params
    _params_['api']._doc = """\
        URL of the GOGS instance without a 'api/<version>' suffix"""

    @staticmethod
    @datasetmethod(name='create_sibling_gogs')
    @eval_results
    def __call__(
            reponame,
            *,
            # possibly retrieve a default from config
            api=None,
            dataset=None,
            recursive=False,
            recursion_limit=None,
            name=None,
            existing='error',
            credential=None,
            access_protocol='https',
            publish_depends=None,
            private=False,
            description=None,
            dry_run=False):

        yield from _create_sibling(
            platform=_GOGS(api, credential, require_token=not dry_run),
            reponame=reponame,
            dataset=dataset,
            recursive=recursive,
            recursion_limit=recursion_limit,
            name=name,
            existing=existing,
            access_protocol=access_protocol,
            publish_depends=publish_depends,
            private=private,
            description=description,
            dry_run=dry_run,
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Common interface options

"""

__docformat__ = 'restructuredtext'

from datalad.interface.results import known_result_xfms
from datalad.support.param import Parameter
from datalad.support.constraints import (
    EnsureBool,
    EnsureCallable,
    EnsureChoice,
    EnsureInt,
    EnsureNone,
    EnsureStr,
    EnsureStrPrefix,
)

location_description = Parameter(
    args=("-D", "--description",),
    constraints=EnsureStr() | EnsureNone(),
    doc="""short description to use for a dataset location. Its primary
    purpose is to help humans to identify a dataset copy (e.g., "mike's dataset
    on lab server"). Note that when a dataset is published, this information
    becomes available on the remote side.""")

recursion_flag = Parameter(
    args=("-r", "--recursive",),
    action="store_true",
    doc="""if set, recurse into potential subdatasets""")

recursion_limit = Parameter(
    args=("-R", "--recursion-limit",),
    metavar="LEVELS",
    constraints=EnsureInt() | EnsureNone(),
    doc="""limit recursion into subdatasets to the given number of levels""")

contains = Parameter(
    args=('--contains',),
    metavar='PATH',
    action='append',
    doc="""limit to the subdatasets containing the
    given path. If a root path of a subdataset is given, the last
    considered dataset will be the subdataset itself.[CMD:  This
    option can be given multiple times CMD][PY:  Can be a list with
    multiple paths PY], in which case datasets that
    contain any of the given paths will be considered.""",
    constraints=EnsureStr() | EnsureNone())

fulfilled = Parameter(
    args=("--fulfilled",),
    doc="""DEPRECATED: use [CMD: --state CMD][PY: `state` PY]
    instead. If given, must be a boolean flag indicating whether
    to consider either only locally present or absent datasets.
    By default all subdatasets are considered regardless of their
    status.""",
    constraints=EnsureBool() | EnsureNone())

dataset_state = Parameter(
    args=("--state",),
    doc="""indicate which (sub)datasets to consider: either only locally present,
    absent, or any of those two kinds.
    """,
    # yoh: intentionally left out the description of default since might be
    # command specific
    constraints=EnsureChoice('present', 'absent', 'any'))

shared_access_opt = Parameter(
    args=('--shared-access',),
    metavar='MODE',
    doc="""configure shared access to a dataset, see `git init --shared`
    documentation for complete details on the supported scenarios. Possible
    values include: 'false', 'true', 'group', and 'all'""")

super_datasets_flag = Parameter(
    args=("-S", "--super-datasets",),
    action="store_true",
    doc="""if set, save a change in a dataset also in its superdataset""")

git_opts = Parameter(
    args=("--git-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git` calls""")

git_clone_opts = Parameter(
    args=("--git-clone-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git clone` calls""")

annex_opts = Parameter(
    args=("--annex-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git annex` calls""")

annex_init_opts = Parameter(
    args=("--annex-init-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git annex init` calls""")

annex_add_opts = Parameter(
    args=("--annex-add-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git annex add` calls""")

annex_get_opts = Parameter(
    args=("--annex-get-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git annex get` calls""")

annex_copy_opts = Parameter(
    args=("--annex-copy-opts",),
    metavar='STRING',
    constraints=EnsureStr() | EnsureNone(),
    doc="""option string to be passed to :command:`git annex copy` calls""")

allow_dirty = Parameter(
    args=("--allow-dirty",),
    action="store_true",
    doc="""flag that operating on a dirty repository (uncommitted or untracked content) is ok""")

if_dirty_opt = Parameter(
    args=("--if-dirty",),
    choices=('fail', 'save-before', 'ignore'),
    doc="""desired behavior if a dataset with unsaved changes is discovered:
    'fail' will trigger an error and further processing is aborted;
    'save-before' will save all changes prior any further action;
    'ignore' let's datalad proceed as if the dataset would not have unsaved
    changes.""")

nosave_opt = Parameter(
    args=("--nosave",),
    dest='save',
    action="store_false",
    doc="""by default all modifications to a dataset are immediately saved. Giving
    this option will disable this behavior.""")

save_message_opt = Parameter(
    args=("-m", "--message",),
    metavar='MESSAGE',
    doc="""a description of the state or the changes made to a dataset.""",
    constraints=EnsureStr() | EnsureNone())

message_file_opt = Parameter(
    args=("-F", "--message-file"),
    doc="""take the commit message from this file. This flag is
    mutually exclusive with -m.""",
    constraints=EnsureStr() | EnsureNone())

reckless_opt = Parameter(
    args=("--reckless",),
    # if no specific mode is given, set to auto
    const='auto',
    nargs='?',
    # boolean types only for backward compatibility
    constraints=
    EnsureChoice(None, True, False, 'auto', 'ephemeral') | \
    EnsureStrPrefix('shared-'),
    metavar='auto|ephemeral|shared-...',
    doc="""Obtain a dataset or subdatset and set it up in a potentially 
    unsafe way for performance, or access reasons. 
    Use with care, any dataset is marked as 'untrusted'.
    The reckless mode is stored in a dataset's local configuration under
    'datalad.clone.reckless', and will be inherited to any of its subdatasets.
    Supported modes are:
    ['auto']: hard-link files between local clones. In-place
    modification in any clone will alter original annex content.
    ['ephemeral']: symlink annex to origin's annex and discard local
    availability info via git-annex-dead 'here' and declares this annex private.
    Shares an annex between origin and clone w/o git-annex being aware of it.
    In case of a change in origin you need to update the clone before you're
    able to save new content on your end.
    Alternative to 'auto' when hardlinks are not an option, or number of consumed
    inodes needs to be minimized. Note that this mode can only be used with clones from
    non-bare repositories or a RIA store! Otherwise two different annex object tree
    structures (dirhashmixed vs dirhashlower) will be used simultaneously, and annex keys
    using the respective other structure will be inaccessible.
    ['shared-<mode>']: set up repository and annex permission to enable multi-user
    access. This disables the standard write protection of annex'ed files.
    <mode> can be any value support by 'git init --shared=', such as 'group', or
    'all'.""")

jobs_opt = Parameter(
    args=("-J", "--jobs"),
    metavar="NJOBS",
    default='auto',
    constraints=EnsureInt() | EnsureNone() | EnsureChoice('auto'),
    doc="""how many parallel jobs (where possible) to use. "auto" corresponds
    to the number defined by 'datalad.runtime.max-annex-jobs' configuration
    item""")

verbose = Parameter(
    args=("-v", "--verbose",),
    action="store_true",
    doc="""print out more detailed information while executing a command""")


as_common_datasrc = Parameter(
    args=("--as-common-datasrc",),
    metavar='NAME',
    doc="""configure the created sibling as a common data source of the
    dataset that can be automatically used by all consumers of the
    dataset (technical: git-annex auto-enabled special remote)""")


publish_depends = Parameter(
    args=("--publish-depends",),
    metavar='SIBLINGNAME',
    doc="""add a dependency such that the given existing sibling is
    always published prior to the new sibling. This equals setting a
    configuration item 'remote.SIBLINGNAME.datalad-publish-depends'.
    [PY: Multiple dependencies can be given as a list of sibling names
    PY][CMD: This option can be given more than once to configure multiple
    dependencies CMD]""",
    action='append',
    constraints=EnsureStr() | EnsureNone())

publish_by_default = Parameter(
    args=("--publish-by-default",),
    metavar='REFSPEC',
    doc="""add a refspec to be published to this sibling by default if nothing
    specified.""",
    constraints=EnsureStr() | EnsureNone(),
    action='append')

annex_wanted_opt = Parameter(
    args=("--annex-wanted",),
    metavar='EXPR',
    doc="""expression to specify 'wanted' content for the repository/sibling.
    See https://git-annex.branchable.com/git-annex-wanted/ for more
    information""",
    constraints=EnsureStr() | EnsureNone())

annex_required_opt = Parameter(
    args=("--annex-required",),
    metavar='EXPR',
    doc="""expression to specify 'required' content for the repository/sibling.
    See https://git-annex.branchable.com/git-annex-required/ for more
    information""",
    constraints=EnsureStr() | EnsureNone())

annex_group_opt = Parameter(
    args=("--annex-group",),
    metavar='EXPR',
    doc="""expression to specify a group for the repository.
    See https://git-annex.branchable.com/git-annex-group/ for more
    information""",
    constraints=EnsureStr() | EnsureNone())

annex_groupwanted_opt = Parameter(
    args=("--annex-groupwanted",),
    metavar='EXPR',
    doc="""expression for the groupwanted.
    Makes sense only if [PY: annex_wanted PY][CMD: --annex-wanted CMD]="groupwanted"
    and annex-group is given too.
    See https://git-annex.branchable.com/git-annex-groupwanted/ for more information""",
    constraints=EnsureStr() | EnsureNone())


inherit_opt = Parameter(
    args=("--inherit",),
    action="store_true",
    doc="""if sibling is missing, inherit settings (git config, git annex
    wanted/group/groupwanted) from its super-dataset""")

missing_sibling_opt = Parameter(
    args=("--missing",),
    constraints=EnsureChoice('fail', 'inherit', 'skip'),  # may be inherit-skip
    metavar='MODE',
    doc="""action to perform, if a sibling does not exist in a given dataset.
    By default it would fail the run ('fail' setting).  With 'inherit' a
    'create-sibling' with '--inherit-settings' will be used to create sibling
    on the remote. With 'skip' - it simply will be skipped.""")

with_plugin_opt = Parameter(
    args=('--with-plugin',),
    nargs='*',
    action='append',
    metavar='PLUGINSPEC',
    doc="""DataLad plugin to run in addition. PLUGINSPEC is a list
    comprised of a plugin name plus optional `key=value` pairs with arguments
    for the plugin call (see `plugin` command documentation for details).
    [PY: PLUGINSPECs must be wrapped in list where each item configures
    one plugin call. Plugins are called in the order defined by this list.
    PY][CMD: This option can be given more than once to run multiple plugins
    in the order in which they are given. CMD]""")

# define parameters to be used by eval_results to tune behavior
# Note: This is done outside eval_results in order to be available when building
# docstrings for the decorated functions
# TODO: May be we want to move them to be part of the classes _params. Depends
# on when and how eval_results actually has to determine the class.
# Alternatively build a callable class with these to even have a fake signature
# that matches the parameters, so they can be evaluated and defined the exact
# same way.

eval_params = dict(
    return_type=Parameter(
        doc="""return value behavior switch. If 'item-or-list' a single
        value is returned instead of a one-item return value list, or a
        list in case of multiple return values. `None` is return in case
        of an empty list.""",
        default='list',
        constraints=EnsureChoice('generator', 'list', 'item-or-list')),
    result_filter=Parameter(
        doc="""if given, each to-be-returned
        status dictionary is passed to this callable, and is only
        returned if the callable's return value does not
        evaluate to False or a ValueError exception is raised. If the given
        callable supports `**kwargs` it will additionally be passed the
        keyword arguments of the original API call.""",
        constraints=EnsureCallable() | EnsureNone()),
    result_xfm=Parameter(
        doc="""if given, each to-be-returned result
        status dictionary is passed to this callable, and its return value
        becomes the result instead. This is different from
        `result_filter`, as it can perform arbitrary transformation of the
        result value. This is mostly useful for top-level command invocations
        that need to provide the results in a particular format. Instead of
        a callable, a label for a pre-crafted result transformation can be
        given.""",
        constraints=EnsureChoice(*list(known_result_xfms.keys())) | EnsureCallable() | EnsureNone()),
    result_renderer=Parameter(
        doc="""select rendering mode command results.
        'tailored' enables a command-specific rendering style that is typically
        tailored to human consumption, if there is one for a specific
        command, or otherwise falls back on the the 'generic' result renderer;
        'generic' renders each result in one line  with key info like action,
        status, path, and an optional message);
        'json' a complete JSON line serialization of the full result record;
        'json_pp' like 'json', but pretty-printed spanning multiple lines;
        'disabled' turns off result rendering entirely;
        '<template>' reports any value(s) of any result properties in any
        format indicated by the template (e.g. '{path}', compare with JSON
        output for all key-value choices). The template syntax follows the
        Python "format() language". It is possible to report individual
        dictionary values, e.g. '{metadata[name]}'. If a 2nd-level key contains
        a colon, e.g. 'music:Genre', ':' must be substituted by '#' in the
        template, like so: '{metadata[music#Genre]}'.""",
        default='tailored'),
    on_failure=Parameter(
        doc="""behavior to perform on failure: 'ignore' any failure is reported,
        but does not cause an exception; 'continue' if any failure occurs an
        exception will be raised at the end, but processing other actions will
        continue for as long as possible; 'stop': processing will stop on first
        failure and an exception is raised. A failure is any result with status
        'impossible' or 'error'. Raised exception is an IncompleteResultsError
        that carries the result dictionaries of the failures in its `failed`
        attribute.""",
        default='continue',
        constraints=EnsureChoice('ignore', 'continue', 'stop')),
)

eval_defaults = {
    k: p.cmd_kwargs.get('default', None)
    for k, p in eval_params.items()
}
"""\
.. deprecated:: 0.16
   This variable will be removed in a future release. The default values for
   all Parameters (possibly overriding by command-specific settings) are now
   available as :class:`Interface` attributes.
"""



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
"""

__docformat__ = 'restructuredtext'

__all__ = ['common_args']

from .helpers import (
    HelpAction,
    LogLevelAction,
)
from datalad.interface.base import eval_params
from datalad.utils import ensure_unicode


_log_level_names = ['critical', 'error', 'warning', 'info', 'debug']

# argument spec template
#<name>=(
#    <id_as_positional>, <id_as_option>
#    {<ArgumentParser.add_arguments_kwargs>}
#)

common_args = dict(
    cfg_overrides=(
        ('-c',),
        dict(action='append',
             dest='cfg_overrides',
             metavar='(:name|name=value)',
             help="""specify configuration setting overrides. They override any
             configuration read from a file. A configuration can also be
             unset temporarily by prefixing its name with a colon (':'), e.g. ':user.name'.
             Overrides specified here may be overridden themselves by
             configuration settings declared as environment variables.
             """)),
    change_path=(
        ('-C',),
        dict(action='append',
             dest='change_path',
             metavar='PATH',
             help="""run as if datalad was started in <path> instead
             of the current working directory.  When multiple -C options are given,
             each subsequent non-absolute -C <path> is interpreted relative to the
             preceding -C <path>.  This option affects the interpretations of the
             path names in that they are made relative to the working directory
             caused by the -C option""")),
    cmd=(
        ('--cmd',),
        dict(dest='_',
             action='store_true',
             help="""syntactical helper that can be used to end the list of
             global command line options before the subcommand label. Options
             taking an arbitrary number of arguments may require to be followed
             by a single --cmd in order to enable identification of the
             subcommand.""")),
    help=(
        ('-h', '--help', '--help-np'),
        dict(nargs=0, action=HelpAction,
             help="""show this help message.  --help-np forcefully disables
                     the use of a pager for displaying the help message""")),
    log_level=(
        ('-l', '--log-level'),
        dict(action=LogLevelAction,
             choices=_log_level_names + [str(x) for x in range(1, 10)],
             metavar="LEVEL",
             default='warning',
             help="""set logging verbosity level.  Choose among %s.  Also you can
             specify an integer <10 to provide even more debugging
             information""" % ', '.join(_log_level_names))),
    # CLI analog of eval_params.on_failure. TODO: dedup
    on_failure=(
        ('--on-failure',),
        dict(dest='common_on_failure',
             # setting the default to None here has the following implications
             # - the global default is solely defined in
             #   datalad.interface.common_opts.eval_params and is in-effect for
             #   Python API and CLI
             # - this global default is written to each command Interface class
             #   and can be overridden there on a per-command basis, with such
             #   override being honored by both APIs
             # - the CLI continues to advertise the choices defined below as
             #   the possible values for '--on-failure'
             # - the Python docstring reflects a possibly command-specific
             #   default
             default=None,
             choices=['ignore', 'continue', 'stop'],
             help="""when an operation fails: 'ignore' and continue with
             remaining operations, the error is logged but does not lead to a
             non-zero exit code of the command; 'continue' works like 'ignore',
             but an error causes a non-zero exit code; 'stop' halts on first
             failure and yields non-zero exit code. A failure is any result
             with status 'impossible' or 'error'. [Default: '%s', but
             individual commands may define an alternative default]"""
             % eval_params['on_failure'].cmd_kwargs['default'])),
    report_status=(
        ('--report-status',),
        dict(dest='common_report_status',
             choices=['success', 'failure', 'ok', 'notneeded', 'impossible',
                      'error'],
             help="""constrain command result report to records matching the
             given status. 'success' is a synonym for 'ok' OR 'notneeded',
             'failure' stands for 'impossible' OR 'error'.""")),
    report_type=(
        ('--report-type',),
        dict(dest='common_report_type',
             choices=['dataset', 'file'],
             action='append',
             help="""constrain command result report to records matching the
             given type. Can be given more than once to match multiple types.
             """)),
    # CLI analog of eval_params.result_renderer but with `<template>` handling
    # and a different default: in Python API we have None as default and do not
    # render the results but return them.  In CLI we default to "default"
    # renderer
    result_renderer=(
        # this should really have --result-renderer for homogeneity with the
        # Python API, but adding it in addition makes the help output
        # monsterous
        ('-f', '--output-format'), # '--result-renderer',
        dict(dest='common_result_renderer',
             default='tailored',
             type=ensure_unicode,
             metavar="{generic,json,json_pp,tailored,disabled,'<template>'}",
             help=eval_params['result_renderer']._doc \
             + " [Default: '%(default)s']")),
)

if __debug__:
    common_args.update(
        dbg=(
            ('--dbg',),
            dict(action='store_true',
                 dest='common_debug',
                 help="enter Python debugger for an uncaught exception",
            )),
        idbg=(
            ('--idbg',),
            dict(action='store_true',
                 dest='common_idebug',
                 help="enter IPython debugger for an uncaught exception")),
    )
